from collections import defaultdict


NODE_POS = 0
NODE_EXPRESSION = 1
MECAB_WORD_FEATURE = 1
IDX_TOKEN = 0

class MeCabStorage:

    """ 메캅 형태소 분석 결과를 원래대로 돌려주는 기능 """

    def __init__(self):
        self.data = defaultdict(list)

    def _append(self, index, sentence):

        """ 인덱스를 키값으로 추가 """

        self.data[index].append(sentence)

    def _mecab_reverse(self):
        reverse_sentence = list()
        for key in sorted(self.data):
            reverse_sentence.append("".join(self.data[key]))
        return reverse_sentence

    def restore_mecab_tokens(self, parse_token):
        for parse_token_item in parse_token:
            self._append(parse_token_item.space_token_idx, parse_token_item.word)
        return self._mecab_reverse()

    def reverse_compound_tokens(self, parse_token):
        """
        복합어로 파싱된 결과를 space_token_idx을 이용해 원래대로 돌려주는 기능
        :param parse_token: 복원할 토큰
        :return: 복원된 문장
        """
        tmp_word = None
        tmp_idx = None

        for parse_token_item in parse_token:
            if parse_token_item[MECAB_WORD_FEATURE].type is None:
                self._append(parse_token_item[MECAB_WORD_FEATURE].space_token_idx, parse_token_item[MECAB_WORD_FEATURE].word)
                tmp_word = None
                continue

            if tmp_word == parse_token_item[MECAB_WORD_FEATURE].word and (
                    tmp_idx == parse_token_item[MECAB_WORD_FEATURE].space_token_idx):
                continue
            self._append(parse_token_item[MECAB_WORD_FEATURE].space_token_idx, parse_token_item[MECAB_WORD_FEATURE].word)
            tmp_word = parse_token_item[MECAB_WORD_FEATURE].word
            tmp_idx = parse_token_item[MECAB_WORD_FEATURE].space_token_idx

        return self._mecab_reverse()


import copy

import _mecab
from collections import namedtuple
from typing import Generator

from mecab import MeCabError
from .domain.mecab_domain import MecabWordFeature


def delete_pattern_from_string(string, pattern, index, nofail=False):
    """ 문자열에서 패턴을 찾아서 *로 변환해주는 기능 """

    # raise an error if index is outside of the string
    if not nofail and index not in range(len(string)):
        raise ValueError("index outside given string")

    # if not erroring, but the index is still not in the correct range..
    if index < 0:  # add it to the beginning
        return pattern + string
    if index > len(string):  # add it to the end
        return string + pattern

    len_pattern = len(pattern)
    blank_pattern = len(pattern) * "*"
    # insert the new string between "slices" of the original
    return string[:index] + blank_pattern + string[index + len_pattern:]

STRING_NOT_FOUND = -1

Feature = namedtuple('Feature', [
    'pos',
    'semantic',
    'has_jongseong',
    'reading',
    'type',
    'start_pos',
    'end_pos',
    'expression',
])


def _create_lattice(sentence):
    lattice = _mecab.Lattice()
    lattice.add_request_type(_mecab.MECAB_ALLOCATE_SENTENCE)  # Required
    lattice.set_sentence(sentence)

    return lattice


def _get_mecab_feature(node) -> MecabWordFeature:
    # Reference:
    # - http://taku910.github.io/mecab/learn.html
    # - https://docs.google.com/spreadsheets/d/1-9blXKjtjeKZqsf4NzHeYJCrr49-nXeRF6D80udfcwY
    # - https://bitbucket.org/eunjeon/mecab-ko-dic/src/master/utils/dictionary/lexicon.py

    # feature = <pos>,<semantic>,<has_jongseong>,<reading>,<type>,<start_pos>,<end_pos>,<expression>
    values = node.feature.split(',')
    assert len(values) == 8

    values = [value if value != '*' else None for value in values]
    feature = dict(zip(Feature._fields, values))
    feature['has_jongseong'] = {'T': True, 'F': False}.get(feature['has_jongseong'])

    return MecabWordFeature(node.surface, **feature)


class MecabParser:

    """
    문장을 형태소 분석하는 클래스.
    형태소 분석시 형태소 분석 토큰, 스페이스 분석 토큰의 인덱스 위치도 함께 저장
    """

    FIRST_WORD = 0
    type_list = ["Compound", "Inflect"]

    def __init__(self, sentence: str, dicpath=''):
        argument = ''

        if dicpath != '':
            argument = '-d %s' % dicpath

        self.tagger = _mecab.Tagger(argument)
        self.sentence = sentence
        self.sentence_token = self.sentence.split()

    def _get_space_token_idx(self, mecab_word_feature: MecabWordFeature) -> int:

        """ 스페이스로 토큰 분석한 인덱스 값 반환 """

        for idx_token, sentence_token_item in enumerate(self.sentence_token):

            index_string = sentence_token_item.find(mecab_word_feature.word)
            if index_string != STRING_NOT_FOUND:

                self.sentence_token[idx_token] = delete_pattern_from_string(sentence_token_item, mecab_word_feature.word, index_string)

                return idx_token

        return False

    def gen_mecab_token_feature(self) -> Generator:

        """
        메캅으로 형태소 분석한 토큰 제너레이터로 반환
        스페이스로 분석한 토큰의 정보와 형태소로 분석한 토큰의 정보 포함
        """

        lattice = _create_lattice(self.sentence)

        if not self.tagger.parse(lattice):
            raise MeCabError(self.tagger.what())

        for mecab_token_idx, mecab_token in enumerate(lattice):
            mecab_token_feature = _get_mecab_feature(mecab_token)
            mecab_token_feature.mecab_token_idx = mecab_token_idx

            space_token_idx = self._get_space_token_idx(mecab_token_feature)

            if space_token_idx is not False:
                mecab_token_feature.space_token_idx = space_token_idx
                mecab_token_feature.word = mecab_token_feature.word.lower()

                yield mecab_token_feature

    def tokenize_mecab_compound(self) -> Generator:

        """
        메캅으로 분석한 토큰 제너레이터로 반환 결과 중에 복합여, 굴절형태소 있는 경우 토큰화
        """

        for compound_include_item in self.gen_mecab_token_feature():
            if compound_include_item.type in self.type_list:
                compound_item_list = compound_include_item.expression.split("+")
                for compound_item in compound_item_list:
                    word, pos_tag, _ = compound_item.split("/")
                    yield word, compound_include_item

            else:
                yield compound_include_item.word, compound_include_item

    def gen_mecab_compound_token_feature(self):
        for idx, x in enumerate(list(self.tokenize_mecab_compound())):
            copy_x = copy.deepcopy(x)
            copy_x[1].mecab_token_compound_idx = idx
            yield copy_x

    def get_word_from_mecab_compound(self, is_list=False):

        """ 메캅 특성에서 단어만 검출"""

        if is_list:
            return [x[self.FIRST_WORD] for x in list(self.gen_mecab_compound_token_feature())]

        return " ".join([x[self.FIRST_WORD] for x in list(self.gen_mecab_compound_token_feature())])


if __name__ == "__main__":

        test_sentence = "나는 서울대병원에 갔어"

        mecab_parse_results = list(
            MecabParser(test_sentence).gen_mecab_token_feature())

        for idx, mecab_parse_item in enumerate(mecab_parse_results):
            print(mecab_parse_item)

        mecab_parse_results = list(
            MecabParser(test_sentence).gen_mecab_compound_token_feature())

        for idx, mecab_parse_item in enumerate(mecab_parse_results):
            print(mecab_parse_item)


import copy
from typing import List, Generator
from pathlib import Path

from .mecab_parser import MecabParser
from .mecab_storage import MeCabStorage
from .mecab_reader import MecabDataReader
from .domain.mecab_domain import MecabWordCategory, Category, MecabPatternData, MecabNerFeature, NerFeature

MECAB_WORD_FEATURE = 0
INFER_FORWARD = 1
INFER_BACKWARD = 2


def find_patterns_idx(pattern, find_tokens: List, parse_character=False) -> List:

    if parse_character:
        pattern = list(pattern)

    if isinstance(pattern, str):
        pattern = pattern.split()

    tmp_save_list = []

    for i in range(len(find_tokens)-len(pattern)+1):
        for j in range(len(pattern)):
            if find_tokens[i+j][MECAB_WORD_FEATURE] != pattern[j]:
                break
        else:
            tmp_save_list.append((i, i+len(pattern)))

    return tmp_save_list








def gen_integrated_entity(blank_list: List) -> Generator:

    """
    1로 채워진 리스트의 시작점과 끝점을 반환
    """

    start_idx = None
    end_idx = None
    is_word_start = False
    for idx, item in enumerate(blank_list):
        if item == MecabNer.FULL_WORD: # 1이 채워진 단어에 대해서 수행
            end_idx = idx # 끝단어를 시작점으로 잡는다.

            if is_word_start is False: # 1로 채워진 단어의 시작이라면, 시작점을 현 인덱스로 저장한다.
                start_idx = idx
                is_word_start = True

            if idx != len(blank_list) - 1: # 인덱스가 끝이 아니라면, 다음 단어를 탐색한다.
                continue

        if (is_word_start is True) and (end_idx is not None):
            yield start_idx, end_idx
            start_idx = None
            end_idx = None
            is_word_start = False # 단어 초기화
            continue


class MecabNer(MecabDataReader):
    """
    엔티티 추출하는 클래스
    - 앞에 단어 품사를 통한 추론 기능
    - 뒤에 단어 품사를 통한 추론 기능
    - 메캅 형태소 분석 비교 및 형태소와 문자열 비교 기능
    """

    MIN_MEANING = 2
    START_IDX = 0
    END_IDX = 1
    ONE_WORD = 1
    MECAB_FEATURE_IDX = 1
    WORD_IDX = 0
    ENTITY = 0
    EMPTY_WORD = 0
    FULL_WORD = 1
    ENTITY_POS_LIST = ["NNG", "NNP", "NNB", "NNBC", "NR", "NP", "XSN", "XR", "SL", "SH", "SN", "UNKNOWN"]

    def __init__(self, ner_path: str = None, search_category: List = None, infer=True):
        super().__init__(ner_path=ner_path, clear_mecab_dir=False)
        self.search_category = search_category
        if search_category is None:
            self.search_category = list(self._get_category_list())
        self.mecab_parsed_list = []
        self.infer = infer

    def _get_category_list(self):
        for path_item in Path(self.mecab_path).iterdir():
            if path_item.suffix == ".txt":
                yield path_item.stem

    def get_category_entity(self) -> Generator:
        """
        문장에서 카테고리별로 엔티티 추출
        :param sentence: 엔티티를 찾고자 하는 문장
        :return: 추출 엔티티
        """

        for mecab_category_item in self.gen_all_mecab_category_data(storage_path=self.mecab_path, use_mecab_parser=False):
            mecab_parsed_copied = copy.deepcopy(self.mecab_parsed_list)
            category, mecab_dictionary_data = mecab_category_item
            if category not in self.search_category:
                continue
            for small_category in mecab_dictionary_data.keys():

                for small_category_item in mecab_dictionary_data.get(small_category):

                    original_data, mecab_data = small_category_item.split(",")

                    category_data = Category(large=category, small=small_category)
                    yield from self.get_pattern(MecabPatternData(category=category_data, dictionary_data=original_data, pattern=mecab_data, sentence=mecab_parsed_copied))
                    yield from self.get_pattern(MecabPatternData(category=category_data, dictionary_data=original_data,pattern=original_data, sentence=mecab_parsed_copied, min_meaning=2, parse_character=True))




    def get_pattern(self, m_p_d: MecabPatternData) -> Generator:
        """
        문장에서 데이터 패턴을 찾는 기능
        - 엔티티의 마지막 단어는 명사형이여야 함. (이가 조사임에도 엔티티로 빠지는 경우 방지)
        :param m_p_d: 찾고자 하는 메캅 데이터 패턴 정보를 담고 있는 클래스
        :return: 문장에서 메캅 단어 정보를 담고 있는 클래스
        """
        space_token_contain_pattern = find_patterns_idx(m_p_d.pattern, m_p_d.sentence, m_p_d.parse_character)

        if (len(m_p_d.pattern) >= m_p_d.min_meaning) and space_token_contain_pattern:
            for pattern_item in space_token_contain_pattern:

                pattern_end_pos = m_p_d.sentence[pattern_item[self.END_IDX] - 1][self.MECAB_FEATURE_IDX].pos

                if pattern_end_pos not in self.ENTITY_POS_LIST:
                    continue

                self._prevent_compound_token(pattern_item, m_p_d.sentence)

                yield MecabWordCategory(category=m_p_d.category,
                                        start_idx=pattern_item[self.START_IDX],
                                        end_idx=pattern_item[self.END_IDX], entity=m_p_d.dictionary_data)

    def infer_entity(self, mecab_parsed_list: List, mecab_category_item: MecabWordCategory) -> MecabWordCategory:
        """ pos에 따라 start_index 변경"""

        end_point = 0
        if mecab_category_item.start_idx == 1:
            end_point = -1

        for idx_range in range(mecab_category_item.start_idx - 1, end_point, -1):
            if mecab_parsed_list[idx_range][MecabNer.MECAB_FEATURE_IDX].pos in ["NNG", "NNP"]:
                mecab_category_item.start_idx = mecab_parsed_list[idx_range][1].mecab_token_compound_idx
                continue
            break
        return mecab_category_item

    def _prevent_compound_token(self, pattern_item: List, mc_p_l: List) -> None:
        """
        - 한 카테고리에서 같은 짧은 단어 추출하는 경우 방지. ex) 양념 치킨 추출 후 치킨 추출
        :param pattern_item: 찾는 패턴
        :param mc_p_l: 문장을 메캅으로 나눈 리스트
        """

        for pattern_idx_item in range(pattern_item[self.START_IDX], pattern_item[self.END_IDX], self.ONE_WORD):
            mc_p_l[pattern_idx_item] = ("*", mc_p_l[pattern_idx_item][self.MECAB_FEATURE_IDX])


    def fill_entity_in_blank(self, mecab_entity_category_list: List) -> List:
        """
        문장에서 인덱스 값이 있는 경우
        :param mecab_entity_category_list:
        :param length:
        :return:
        """
        blank = [MecabNer.EMPTY_WORD] * len(self.mecab_parsed_list)

        for mecab_entity_category_item in mecab_entity_category_list:
            for i in range(mecab_entity_category_item.start_idx, mecab_entity_category_item.end_idx, 1):
                blank[i] = self.FULL_WORD

        return blank

    def gen_mecab_category_entity(self):


        mecab_entity_category_list = []
        for category_entity_item in self.get_category_entity():
            mecab_entity_category_list.append(self.infer_entity(self.mecab_parsed_list, category_entity_item))

        many_entity_index_list = self.fill_entity_in_blank(mecab_entity_category_list)

        for integrated_entity_item in gen_integrated_entity(many_entity_index_list):
            end_idx = integrated_entity_item[1] + 1
            start_idx = integrated_entity_item[0]
            mecab_parsed_token = self.mecab_parsed_list[start_idx:end_idx]
            restore_tokens = MeCabStorage().reverse_compound_tokens(mecab_parsed_token)
            restore_sentence = " ".join(restore_tokens)
            for entity_category_item in mecab_entity_category_list:
                if entity_category_item.end_idx == end_idx:
                    small_category_replace = entity_category_item.category.small.replace("#", "").strip()
                    yield MecabNerFeature(word=restore_sentence,
                                        pos="ner",
                                        start_idx=start_idx,
                                        end_idx=end_idx,
                                        category=Category(large=entity_category_item.category.large, small=small_category_replace))

    def parse(self, sentence: str):
        self.mecab_parsed_list = list(MecabParser(sentence=sentence).gen_mecab_compound_token_feature())

        mecab_cat_list = list(self.gen_mecab_category_entity())
        cat_idx_list=[]
        [cat_idx_list.extend(list(range(x.start_idx, x.end_idx, 1))) for x in mecab_cat_list]

        parse_result = []
        for idx, mecab_parse_item in enumerate(self.mecab_parsed_list):
            for mecab_item in mecab_cat_list:
                if idx == mecab_item.end_idx:
                    parse_result.append((mecab_item.word, mecab_item))
                    break
            if idx in cat_idx_list:
                continue
            parse_result.append((mecab_parse_item[1].word, NerFeature(word=mecab_parse_item[self.MECAB_FEATURE_IDX].word, pos=mecab_parse_item[self.MECAB_FEATURE_IDX].pos)))
        return parse_result

    def morphs(self, sentence: str):
        return [x[self.WORD_IDX] for x in self.parse(sentence=sentence)]

    def ners(self, sentence: str):
        result = [(x[self.MECAB_FEATURE_IDX].word, x[self.MECAB_FEATURE_IDX].category.large, x[self.MECAB_FEATURE_IDX].category.small) for x in self.parse(sentence=sentence) if x[self.MECAB_FEATURE_IDX].pos == "ner"]
        return result


from typing import List, Generator
from pathlib import Path
from collections import defaultdict

from .domain.mecab_exception import MecabDataReaderException
from .mecab_parser import MecabParser


class CategoryData:
    def __init__(self):
        self.data = defaultdict(list)

    def add(self, header, data):
        self.data[header].extend(data)


class MecabDataReader:
    FIRST_WORD = 0
    MECAB_DATA = "mecab_data"

    FORMAT_SUFFIX = ".txt"
    HEADER = "#"

    def __init__(self, ner_path: str = None, clear_mecab_dir=False):
        self.ner_path = ner_path or "./"
        self._set_mecab_path(self.ner_path, clear_mecab_dir)

    def _set_mecab_path(self, path, clear_dir) -> None:
        """ 경로 검증 """
        if not Path(path).is_dir():
            raise MecabDataReaderException("Please check if directory is proper")

        self.mecab_path = Path(__file__).parent.joinpath("data", MecabDataWriter.MECAB_DATA)

        if clear_dir:
            self._clear_dir()

        if not Path(self.mecab_path).exists():
            Path(self.mecab_path).mkdir()

    def _clear_dir(self):
        """메캅 관련 디렉터리 전부 삭제하는 메소드"""
        try:
            for path_item in Path(self.mecab_path).iterdir():
                Path(path_item).unlink()
            Path(self.mecab_path).rmdir()
        except FileNotFoundError:
            pass

    @classmethod
    def read_category(cls, txt_data: List) -> Generator:
        """
        데이터에서 헤더, 내용을 나눠서 반환하는 메소드
        헤더로 시작되지 않을시 리스트로 반환
        """
        header, *contents = txt_data
        category_list = []

        for data_item in contents:
            if cls.HEADER in data_item:
                yield header, sorted(category_list, key=len, reverse=True)
                header = data_item
                category_list = []
                continue

            if data_item == '':
                continue

            category_list.append(data_item)

        yield header, sorted(category_list, key=len, reverse=True)

    @classmethod
    def gen_all_mecab_category_data(cls, storage_path, use_mecab_parser=False) -> Generator:
        """경로에 있는 데이터 읽은 뒤 카테고리 데이터셋으로 반환"""

        for path_item in Path(storage_path).iterdir():
            if Path(path_item).suffix != cls.FORMAT_SUFFIX:
                continue
            txt_data = cls.read_txt(path_item)
            c_d = CategoryData()

            if not txt_data[cls.FIRST_WORD].startswith(cls.HEADER):
                txt_data.insert(cls.FIRST_WORD, cls.HEADER + path_item.stem)

            for data_item in cls.read_category(txt_data):

                data_header, contents = data_item

                if use_mecab_parser:
                    contents = [(x, MecabParser(sentence=x).get_word_from_mecab_compound()) for x in contents]

                c_d.add(data_header, contents)

            yield path_item.stem, dict(c_d.data)

    @staticmethod
    def read_txt(path):
        """텍스트 파일 읽는 메소드"""
        with open(path, "r", encoding='utf-8-sig') as file:
            txt_list = file.read().splitlines()
            return txt_list


class MecabDataWriter(MecabDataReader):
    MECAB_WORD = 1
    ORIGIN_WORD = 0

    ITEM_BOUNDARY = ","
    CATEGORY_SPLITER = "_"



    def write_category(self) -> None:
        """카테고리별로 데이터 저장하는 메소드"""

        for data_item in self.gen_all_mecab_category_data(storage_path=self.ner_path, use_mecab_parser=True):
            category, content = data_item

            file_name = category + self.FORMAT_SUFFIX
            mecab_write_path = self.mecab_path.joinpath(file_name)

            for content_key_item in content.keys():
                mecab_write_list = [content_key_item,]
                mecab_write_list.extend([str(x[self.ORIGIN_WORD]) + self.ITEM_BOUNDARY + str(x[self.MECAB_WORD]) for x in
                             content[content_key_item]])
                MecabDataWriter.write_txt(path=str(mecab_write_path), txt_list=mecab_write_list)


    @staticmethod
    def write_txt(path: str, txt_list: List, is_sort=False):
        """텍스트 파일 쓰는 메소드"""
        if is_sort:
            txt_list = sorted(list(txt_list), key=len, reverse=True)

        with open(path, "a", encoding='UTF8') as file:
            for idx, txt_item in enumerate(txt_list):
                data = str(txt_item) + "\n"
                file.write(data)